#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import argparse
import json
import logging
import os

from scraper import code_gov
from scraper.util import configure_logging

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        description="Scrape code repositories for Code.gov / DOE CODE"
    )

    parser.add_argument(
        "--agency", type=str, nargs="?", default="", help='Agency Label, e.g. "DOE"'
    )
    parser.add_argument(
        "--method",
        type=str,
        nargs="?",
        default="",
        help="Method of measuring open source",
    )
    parser.add_argument(
        "--organization",
        type=str,
        nargs="?",
        default="",
        help="Force all repos to report a particular organization",
    )
    parser.add_argument(
        "--contact-email",
        type=str,
        nargs="?",
        default="",
        help="Force all repos to report a particular contact email",
    )

    parser.add_argument(
        "--config", type=str, nargs="?", default="", help="Configuration File (*.json)"
    )

    parser.add_argument(
        "--github-gov-orgs",
        action="store_true",
        help="Use orgs from government.github.com/community",
    )
    parser.add_argument(
        "--skip-labor-hours",
        action="store_true",
        help='Skip calculation of labor hours, assume "0"',
    )

    parser.add_argument(
        "--doecode-json",
        type=str,
        nargs="?",
        default=None,
        help="Path to DOE CODE .json file",
    )
    parser.add_argument(
        "--doecode-url",
        type=str,
        nargs="?",
        default=None,
        help="URL to DOE CODE .json data",
    )
    parser.add_argument(
        "--doecode-api-key",
        type=str,
        nargs="?",
        default=None,
        help="DOE CODE API key for accessing --doecode-url",
    )

    parser.add_argument(
        "--output-path",
        type=str,
        nargs="?",
        default="",
        help="Output path for .json file",
    )
    parser.add_argument(
        "--output-filename",
        type=str,
        nargs="?",
        default="code.json",
        help="Output filename for .json file",
    )

    parser.add_argument("--verbose", action="store_true", help="Enable verbose output")

    args = parser.parse_args()

    configure_logging(args.verbose)

    try:
        config_json = json.load(open(args.config))
    except (FileNotFoundError, json.JSONDecodeError):
        if args.config:
            raise
        config_json = {}

    # Update config based on commandline arguments
    if args.agency:
        config_json["agency"] = args.agency
    if args.method:
        config_json["method"] = args.method
    if args.organization:
        config_json["organization"] = args.organization
    if args.contact_email:
        config_json["contact_email"] = args.contact_email
    if args.output_path:
        config_json["output_path"] = args.output_path
    if args.skip_labor_hours:
        config_json["compute_labor_hours"] = False
    if args.github_gov_orgs:
        config_json["github_gov_orgs"] = True

    config_json["DOE CODE"] = {}
    config_json["DOE CODE"]["json"] = args.doecode_json
    config_json["DOE CODE"]["url"] = args.doecode_url
    config_json["DOE CODE"]["api_key"] = args.doecode_api_key

    output_path = config_json.get("output_path", None)
    output_path = args.output_path or output_path
    logger.debug("Output Path: %s", output_path)

    if output_path is not None and not os.path.exists(output_path):
        raise RuntimeError(
            "Invalid output path argument provided!  Make sure the output path exists and try again."
        )

    code_json = code_gov.process_config(config_json)

    code_gov.force_attributes(code_json, config_json)

    logger.info("Number of Projects: %s", len(code_json["releases"]))

    output_filepath = args.output_filename

    if output_path is not None:
        output_filepath = os.path.join(output_path, output_filepath)

    with open(output_filepath, "w", encoding="utf-8") as fp:
        logger.info("Writing output to: %s", output_filepath)
        fp.write(code_json.to_json())


if __name__ == "__main__":
    main()



import json
import logging
import logging.config
import os
import re
import requests
import tempfile

from subprocess import Popen, PIPE, STDOUT  # nosec

logger = logging.getLogger(__name__)

EFFORT_REGEX = re.compile(r"Effort = ([\d\.]+) Person-months")


def execute(command, cwd=None):
    logger.debug("Forking command: %s", command)

    if cwd is None:
        cwd = os.getcwd()
    elif not os.path.isdir(cwd):
        raise ValueError("path does not exist: %s", cwd)

    process = Popen(command, cwd=cwd, stdout=PIPE, stderr=STDOUT, shell=False)  # nosec
    out, err = process.communicate()
    return str(out), str(err)


def configure_logging(verbose=False):
    DEFAULT_LOGGING = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {
                # 'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
                # 'format': '%(levelname)s: %(message)s'
                "format": "%(asctime)s - %(levelname)s: %(message)s"
            }
        },
        "handlers": {
            "default": {
                "level": "INFO",
                "formatter": "standard",
                "class": "logging.StreamHandler",
            },
            "null": {
                "level": "INFO",
                "formatter": "standard",
                "class": "logging.NullHandler",
            },
        },
        "loggers": {
            "": {"handlers": ["default"], "level": "DEBUG", "propagate": False},
            "github3": {"handlers": ["null"], "level": "DEBUG", "propagate": False},
            "urllib3": {"handlers": ["null"], "level": "DEBUG", "propagate": False},
        },
    }

    if verbose:
        DEFAULT_LOGGING["handlers"]["default"]["level"] = "DEBUG"
        # DEFAULT_LOGGING['loggers']['']['level'] = 'DEBUG'

    logging.config.dictConfig(DEFAULT_LOGGING)


def git_repo_to_sloc(url):
    """
    Given a Git repository URL, returns number of lines of code based on cloc

    Reference:
    - cloc: https://github.com/AlDanial/cloc
    - https://www.omg.org/spec/AFP/
        - Another potential way to calculation effort

    Sample cloc output:
        {
            "header": {
                "cloc_url": "github.com/AlDanial/cloc",
                "cloc_version": "1.74",
                "elapsed_seconds": 0.195950984954834,
                "n_files": 27,
                "n_lines": 2435,
                "files_per_second": 137.78956000769,
                "lines_per_second": 12426.5769858787
            },
            "C++": {
                "nFiles": 7,
                "blank": 121,
                "comment": 314,
                "code": 371
            },
            "C/C++ Header": {
                "nFiles": 8,
                "blank": 107,
                "comment": 604,
                "code": 191
            },
            "CMake": {
                "nFiles": 11,
                "blank": 49,
                "comment": 465,
                "code": 165
            },
            "Markdown": {
                "nFiles": 1,
                "blank": 18,
                "comment": 0,
                "code": 30
            },
            "SUM": {
                "blank": 295,
                "comment": 1383,
                "code": 757,
                "nFiles": 27
            }
        }
    """

    with tempfile.TemporaryDirectory() as tmp_dir:
        logger.debug("Cloning: url=%s tmp_dir=%s", url, tmp_dir)

        tmp_clone = os.path.join(tmp_dir, "clone-dir")

        cmd = ["git", "clone", "--depth=1", url, tmp_clone]
        execute(cmd)

        cmd = ["cloc", "--json", tmp_clone]
        out, _ = execute(cmd)

        try:
            json_start = out.find('{"header"')
            json_blob = out[json_start:].replace("\\n", "").replace("'", "")
            cloc_json = json.loads(json_blob)
            sloc = cloc_json["SUM"]["code"]
        except json.decoder.JSONDecodeError:
            logger.debug("Error Decoding: url=%s, out=%s", url, out)
            sloc = 0

    logger.debug("SLOC: url=%s, sloc=%d", url, sloc)

    return sloc


def compute_labor_hours(sloc, month_hours="cocomo_book"):
    """
    Compute the labor hours, given a count of source lines of code

    The intention is to use the COCOMO II model to compute this value.

    References:
    - https://csse.usc.edu/tools/cocomoii.php
    - http://docs.python-guide.org/en/latest/scenarios/scrape/
    """
    # Calculation of hours in a month
    if month_hours == "hours_per_year":
        # Use number of working hours in a year:
        # (40 Hours / week) * (52 weeks / year) / (12 months / year) ~= 173.33
        HOURS_PER_PERSON_MONTH = 40.0 * 52 / 12
    else:
        # Use value from COCOMO II Book (month_hours=='cocomo_book'):
        # Reference: https://dl.acm.org/citation.cfm?id=557000
        # This is the value used by the Code.gov team:
        # https://github.com/GSA/code-gov/blob/master/LABOR_HOUR_CALC.md
        HOURS_PER_PERSON_MONTH = 152.0

    cocomo_url = "https://csse.usc.edu/tools/cocomoii.php"
    page = requests.post(cocomo_url, data={"new_size": sloc})

    try:
        person_months = float(EFFORT_REGEX.search(page.text).group(1))
    except AttributeError:
        logger.error("Unable to find Person Months in page text: sloc=%s", sloc)
        # If there is no match, and .search(..) returns None
        person_months = 0

    labor_hours = person_months * HOURS_PER_PERSON_MONTH
    logger.debug("sloc=%d labor_hours=%d", sloc, labor_hours)

    return labor_hours


def labor_hours_from_url(url):
    sum_sloc = git_repo_to_sloc(url)
    logger.info("SLOC: %d", sum_sloc)

    labor_hours = compute_labor_hours(sum_sloc)
    logger.info("labor_hours: %d", labor_hours)

    return labor_hours


def _prune_dict_null_str(dictionary):
    """
    Prune the "None" or emptry string values from dictionary items
    """
    for key, value in list(dictionary.items()):
        if value is None or str(value) == "":
            del dictionary[key]

        if isinstance(value, dict):
            dictionary[key] = _prune_dict_null_str(dictionary[key])

    return dictionary



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import logging

from scraper.code_gov.models import Metadata, Project
from scraper.github import gov_orgs
from scraper import github, gitlab, bitbucket, doecode, tfs

logger = logging.getLogger(__name__)


def process_config(config):
    """
    Master function to process a Scraper config file

    Returns a Code.gov Metadata file
    """

    agency = config.get("agency", "UNKNOWN")
    logger.debug("Agency: %s", agency)

    method = config.get("method", "other")
    logger.debug("Inventory Method: %s", method)

    compute_labor_hours = config.get("compute_labor_hours", True)

    if config.get("contact_email", None) is None:
        # A default contact email is required to handle the (frequent) case
        # where a project / repository has no available contact email.
        logger.warning('Config file should contain a "contact_email"')

    logger.debug("Creating inventory from config: %s", config)
    code_gov_metadata = Metadata(agency, method)

    # Parse config for GitHub repositories
    github_instances = config.get("GitHub", [])
    if config.get("github_gov_orgs", False):
        github_instances.append({"url": "https://github.com", "orgs": gov_orgs()})
    for instance in github_instances:
        url = instance.get("url", "https://github.com")
        orgs = instance.get("orgs", [])
        repos = instance.get("repos", [])
        public_only = instance.get("public_only", True)
        excluded = instance.get("exclude", [])
        token = instance.get("token", None)

        gh_session = github.connect(url, token)

        for repo in github.query_repos(gh_session, orgs, repos, public_only):
            if repo.owner.login in excluded or repo.full_name in excluded:
                logger.info("Excluding: %s", repo.full_name)
                continue

            code_gov_project = Project.from_github3(
                repo, labor_hours=compute_labor_hours
            )
            code_gov_metadata["releases"].append(code_gov_project)

    # Parse config for GitLab repositories
    gitlab_instances = config.get("GitLab", [])
    for instance in gitlab_instances:
        url = instance.get("url")
        # orgs = instance.get('orgs', [])
        repos = instance.get("repos", [])
        # public_only = instance.get('public_only', True)
        excluded = instance.get("exclude", [])
        token = instance.get("token", None)
        fetch_languages = instance.get("fetch_languages", False)

        gl_session = gitlab.connect(url, token)

        for repo in gitlab.query_repos(gl_session, repos):
            namespace = repo.namespace["path"]
            path_with_namespace = repo.path_with_namespace
            if namespace in excluded or path_with_namespace in excluded:
                logger.info("Excluding: %s", repo.path_with_namespace)
                continue

            code_gov_project = Project.from_gitlab(
                repo, labor_hours=compute_labor_hours, fetch_languages=fetch_languages
            )
            code_gov_metadata["releases"].append(code_gov_project)

    # Parse config for Bitbucket repositories
    bitbucket_instances = config.get("Bitbucket", [])
    for instance in bitbucket_instances:
        url = instance.get("url")
        # orgs = instance.get('orgs', None)
        # public_only = instance.get('public_only', True)
        username = instance.get("username", None)
        password = instance.get("password", None)
        token = instance.get("token", None)
        excluded = instance.get("exclude", [])

        bb_session = bitbucket.connect(url, username, password, token)

        for repo in bitbucket.all_repos(bb_session):
            project = repo["project"]["key"]
            project_repo = "%s/%s" % (project, repo["slug"])
            if project in excluded or project_repo in excluded:
                logger.info("Excluding: %s", project_repo)
                continue

            code_gov_project = Project.from_stashy(
                repo, labor_hours=compute_labor_hours
            )
            code_gov_metadata["releases"].append(code_gov_project)

    # Parse config for TFS repositories
    tfs_instances = config.get("TFS", [])
    for instance in tfs_instances:
        url = instance.get("url")
        token = instance.get("token", None)

        projects = tfs.get_projects_metadata(url, token)
        for project in projects:
            code_gov_project = Project.from_tfs(
                project, labor_hours=compute_labor_hours
            )
            code_gov_metadata["releases"].append(code_gov_project)

    # Handle parsing of DOE CODE records

    doecode_config = config.get("DOE CODE", {})
    doecode_json = doecode_config.get("json", None)
    doecode_url = doecode_config.get("url", None)
    doecode_key = doecode_config.get("api_key", None)

    for record in doecode.process(doecode_json, doecode_url, doecode_key):
        code_gov_project = Project.from_doecode(record)
        code_gov_metadata["releases"].append(code_gov_project)

    return code_gov_metadata


def force_attributes(metadata, config):
    """
    Forces certain fields in the Code.gov Metadata json
    """

    organization = config.get("organization", "")
    logger.debug("Organization: %s", organization)

    contact_email = config.get("contact_email")
    logger.debug("Contact Email: %s", contact_email)

    permissions = config.get("permissions", {})
    default_usage = permissions.get("usageType", "")
    default_exemption_text = permissions.get("exemptionText", "")
    logger.debug("Default usageType: %s", default_usage)
    logger.debug("Default exemptionText: %s", default_exemption_text)

    # Force certain fields
    if organization:
        logger.debug("Forcing Organization to: %s", organization)

    if contact_email:
        logger.debug("Forcing Contact Email to: %s", contact_email)

    for release in metadata["releases"]:
        if organization:
            release["organization"] = organization

        if contact_email:
            release["contact"]["email"] = contact_email

        if "licenses" not in release["permissions"]:
            release["permissions"]["licenses"] = None

        if "description" not in release:
            release["description"] = "No description available..."

        if "usageType" not in release["permissions"]:
            release["permissions"]["usageType"] = default_usage
            release["permissions"]["exemptionText"] = default_exemption_text

    return metadata



import logging
import stashy

import datetime

from stashy.client import Stash

logger = logging.getLogger(__name__)


def connect(url, username=None, password=None, token=None):
    """
    Return a connected Bitbucket session
    """
    if token is not None:
        bb_session = Stash(url, token=token)
        logger.info("Connected to: %s with token", url)
    else:
        bb_session = stashy.connect(url, username, password)
        logger.info("Connected to: %s as username %s", url, username)

    return bb_session


def all_repos(bb_session):
    """
    Yields Stashy repo dictionary objects for all repos in Bitbucket
    """

    for repo in bb_session.repos.all():
        all_commits = sorted(
            bb_session.projects[repo["project"]["key"]]
            .repos[repo["name"]]
            .commits(None),
            key=lambda x: x["authorTimestamp"],
        )
        if all_commits:
            repo["created"] = (
                datetime.datetime.fromtimestamp(
                    all_commits[0]["authorTimestamp"] / 1000
                )
                .date()
                .isoformat()
            )
            repo["lastModified"] = (
                datetime.datetime.fromtimestamp(
                    all_commits[-1]["authorTimestamp"] / 1000
                )
                .date()
                .isoformat()
            )
        yield repo



#! /usr/bin/env python3

import argparse
import json

from scraper.util import compute_labor_hours, git_repo_to_sloc


parser = argparse.ArgumentParser(
    description="Scrape code repositories for Code.gov / DOECode"
)
parser.add_argument(
    "filename", type=str, help="Path to locally stored `code.json` file"
)
args = parser.parse_args()

code_gov_json = json.load(open(args.filename))
releases = code_gov_json["releases"]

repo_urls = {
    release["repositoryURL"].rstrip("/")
    for release in releases
    if release.get("vcs", "") == "git"
}

for url in repo_urls:
    # print(url)

    sloc = git_repo_to_sloc(url)
    # print(sloc)

    hours = compute_labor_hours(sloc)
    print("-- url=%s, sloc=%d, hours=%d" % (url, sloc, hours))



import logging

logger = logging.getLogger(__name__)


def _license_obj(license):
    """
    A helper function to look up license object information

    Use names from: https://api.github.com/licenses
    """
    obj = None

    if license in ("MIT", "MIT License"):
        obj = {"URL": "https://api.github.com/licenses/mit", "name": "MIT"}
    elif license in ('BSD 2-clause "Simplified" License'):
        obj = {
            "URL": "https://api.github.com/licenses/bsd-2-clause",
            "name": "BSD-2-Clause",
        }
    elif license in ('BSD 3-clause "New" or "Revised" License'):
        obj = {
            "URL": "https://api.github.com/licenses/bsd-3-clause",
            "name": "BSD-3-Clause",
        }
    elif license in ("Apache License 2.0"):
        obj = {
            "URL": "https://api.github.com/licenses/apache-2.0",
            "name": "Apache-2.0",
        }
    elif license in ("GNU General Public License v2.1"):
        obj = {"URL": "https://api.github.com/licenses/gpl-2.1", "name": "GPL-2.1"}
    elif license in ("GNU General Public License v2.0"):
        obj = {"URL": "https://api.github.com/licenses/gpl-2.0", "name": "GPL-2.0"}
    elif license in ("GNU Lesser General Public License v2.1"):
        obj = {"URL": "https://api.github.com/licenses/lgpl-2.1", "name": "LGPL-2.1"}
    elif license in ("GNU General Public License v3.0"):
        obj = {"URL": "https://api.github.com/licenses/gpl-3.0", "name": "GPL-3.0"}
    elif license in ("GNU Lesser General Public License v3.0"):
        obj = {"URL": "https://api.github.com/licenses/lgpl-3.0", "name": "LGPL-3.0"}
    elif license in ("Eclipse Public License 1.0"):
        obj = {"URL": "https://api.github.com/licenses/epl-1.0", "name": "EPL-1.0"}
    elif license in ("Mozilla Public License 2.0"):
        obj = {"URL": "https://api.github.com/licenses/mpl-2.0", "name": "MPL-2.0"}
    elif license in ("The Unlicense"):
        obj = {"URL": "https://api.github.com/licenses/unlicense", "name": "Unlicense"}
    elif license in ("GNU Affero General Public License v3.0"):
        obj = {"URL": "https://api.github.com/licenses/agpl-3.0", "name": "AGPL-3.0"}
    elif license in ("Eclipse Public License 2.0"):
        obj = {"URL": "https://api.github.com/licenses/epl-2.0", "name": "EPL-2.0"}

    if obj is None:
        logger.warn("I don't understand the license: %s", license)
        raise ValueError("Aborting!")

    return obj



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.md") as fh:
    long_description = fh.read()

with open("requirements/production.txt") as fp:
    lines = [x.strip() for x in fp.readlines() if x]
    install_reqs = [x for x in lines if not x.startswith("#")]

setup(
    name="llnl-scraper",
    version="0.9.0",
    description="Package for extracting software repository metadata",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ian Lee",
    author_email="lee1001@llnl.gov",
    url="https://github.com/llnl/scraper",
    packages=find_packages(),
    install_requires=install_reqs,
    entry_points={"console_scripts": ["scraper = scraper.gen_code_gov_json:main",]},
    scripts=["scripts/codegov_compute_hours.py",],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)
